@@ -1,12 +1,9 @@
-/* $Id: reqs.c,v 1.82 2002/06/06 20:32:30 rjkaes Exp $
+/* $Id: reqs.c,v 1.105 2003/06/26 18:19:57 rjkaes Exp $
  *
  * This is where all the work in tinyproxy is actually done. Incoming
  * connections have a new child created for them. The child then
  * processes the headers from the client, the response from the server,
  * and then relays the bytes between the two.
- * If TUNNEL_SUPPORT is enabled, then tinyproxy will actually work
- * as a simple buffering TCP tunnel. Very cool! (Robert actually uses
- * this feature for a buffering NNTP tunnel.)
  *
  * Copyright (C) 1998	    Steven Young
  * Copyright (C) 1999-2002  Robert James Kaes (rjkaes@flarenet.com)
@@ -33,6 +30,7 @@
 #include "filter.h"
 #include "hashmap.h"
 #include "heap.h"
+#include "htmlerror.h"
 #include "log.h"
 #include "network.h"
 #include "regexp.h"
@@ -43,32 +41,27 @@
 #include "utils.h"
 #include "vector.h"
 
-#define HTTP400ERROR "Unrecognizable request. Only HTTP is allowed."
-#define HTTP500ERROR "Unable to connect to remote server."
-#define HTTP503ERROR "Internal server error."
-
 /*
  * Maximum length of a HTTP line
  */
 #define HTTP_LINE_LENGTH (MAXBUFFSIZE / 6)
 
+/*
+ * Port constants for HTTP (80) and SSL (443)
+ */
+#define HTTP_PORT 80
+#define HTTP_PORT_SSL 443
+
 /*
  * Macro to help test if the Upstream proxy supported is compiled in and
  * enabled.
  */
 #ifdef UPSTREAM_SUPPORT
-#  define UPSTREAM_CONFIGURED() (config.upstream_name && config.upstream_port != -1)
+#  define UPSTREAM_CONFIGURED() (config.upstream_list != NULL)
+#  define UPSTREAM_HOST(host) upstream_get(host)
 #else
 #  define UPSTREAM_CONFIGURED() (0)
-#endif
-
-/*
- * Macro to help test if tunnel support is compiled in, and is enabled.
- */
-#ifdef TUNNEL_SUPPORT
-#  define TUNNEL_CONFIGURED() (config.tunnel_name && config.tunnel_port != -1)
-#else
-#  define TUNNEL_CONFIGURED() (0)
+#  define UPSTREAM_HOST(host) (NULL)
 #endif
 
 /*
@@ -82,6 +75,19 @@
  */
 static vector_t ports_allowed_by_connect = NULL;
 
+/*
+ * This structure holds the information pulled from a URL request.
+ */
+struct request_s {
+	char *method;
+	char *protocol;
+
+	char *host;
+	uint16_t port;
+
+	char *path;
+};
+
 /*
  * Now, this routine adds a "port" to the list.  It also creates the list if
  * it hasn't already by done.
@@ -98,7 +104,7 @@ add_connect_port_allowed(int port)
 	}
 
 	log_message(LOG_INFO, "Adding Port [%d] to the list allowed by CONNECT", port);
-	vector_insert(ports_allowed_by_connect, (void **)&port, sizeof(port));
+	vector_append(ports_allowed_by_connect, (void **)&port, sizeof(port));
 }
 
 /*
@@ -121,8 +127,9 @@ check_allowed_connect_ports(int port)
 	if (!ports_allowed_by_connect)
 		return 1;
 
-	for (i = 0; i < vector_length(ports_allowed_by_connect); ++i) {
-		if (vector_getentry(ports_allowed_by_connect, i, (void **)&data) < 0)
+	for (i = 0; i != vector_length(ports_allowed_by_connect); ++i) {
+		data = vector_getentry(ports_allowed_by_connect, i, NULL);
+		if (!data)
 			return -1;
 
 		if (*data == port)
@@ -172,18 +179,8 @@ read_request_line(struct conn_s *connptr)
 }
 
 /*
- * This structure holds the information pulled from a URL request.
+ * Free all the memory allocated in a request.
  */
-struct request_s {
-	char *method;
-	char *protocol;
-
-	char *host;
-	uint16_t port;
-
-	char *path;
-};
-
 static void
 free_request_struct(struct request_s *request)
 {
@@ -202,7 +199,28 @@ free_request_struct(struct request_s *request)
 }
 
 /*
- * Pull the information out of the URL line.
+ * Take a host string and if there is a username/password part, strip
+ * it off.
+ */
+static void
+strip_username_password(char* host)
+{
+	char *ptr1, *ptr2;
+
+	if ((ptr1 = strchr(host, '@')) != NULL) {
+		ptr1++; /* move to one past the @ symbol */
+		ptr2 = host;
+
+		/* copy the bytes up to the NUL */
+		while (*ptr1) 
+			*ptr2++ = *ptr1++;
+		*ptr2 = '\0';
+	}
+}
+
+/*
+ * Pull the information out of the URL line.  This will handle both HTTP
+ * and FTP (proxied) URLs.
  */
 static int
 extract_http_url(const char *url, struct request_s *request)
@@ -214,21 +232,24 @@ extract_http_url(const char *url, struct request_s *request)
 		goto ERROR_EXIT;
 
 	if (sscanf
-	    (url, "http://%[^:/]:%hu%s", request->host, &request->port,
+	    (url, "%[^:/]:%hu%s", request->host, &request->port,
 	     request->path) == 3) ;
-	else if (sscanf(url, "http://%[^/]%s", request->host, request->path) == 2)
-		request->port = 80;
-	else if (sscanf(url, "http://%[^:/]:%hu", request->host, &request->port)
+	else if (sscanf(url, "%[^/]%s", request->host, request->path) == 2)
+		request->port = HTTP_PORT;
+	else if (sscanf(url, "%[^:/]:%hu", request->host, &request->port)
 		 == 2)
 		strcpy(request->path, "/");
-	else if (sscanf(url, "http://%[^/]", request->host) == 1) {
-		request->port = 80;
+	else if (sscanf(url, "%[^/]", request->host) == 1) {
+		request->port = HTTP_PORT;
 		strcpy(request->path, "/");
 	} else {
 		log_message(LOG_ERR, "extract_http_url: Can't parse URL.");
 		goto ERROR_EXIT;
 	}
 
+	/* Remove the username/password if they're present */
+	strip_username_password(request->host);
+
 	return 0;
 
   ERROR_EXIT:
@@ -252,7 +273,7 @@ extract_ssl_url(const char *url, struct request_s *request)
 
 	if (sscanf(url, "%[^:]:%hu", request->host, &request->port) == 2) ;
 	else if (sscanf(url, "%s", request->host) == 1)
-		request->port = 443;
+		request->port = HTTP_PORT_SSL;
 	else {
 		log_message(LOG_ERR, "extract_ssl_url: Can't parse URL.");
 
@@ -260,20 +281,210 @@ extract_ssl_url(const char *url, struct request_s *request)
 		return -1;
 	}
 
+	/* Remove the username/password if they're present */
+	strip_username_password(request->host);
+
 	return 0;
 }
 
+#ifdef TRANSPARENT_PROXY
+/*
+ * Build a URL from parts.
+ */
+static int
+build_url(char **url, const char *host, int port, const char *path)
+{
+	int len;
+
+	assert(url != NULL);
+	assert(host != NULL);
+	assert(port > 0 && port < 32768);
+	assert(path != NULL);
+
+	len = strlen(host) + strlen(path) + 14;
+	*url = safemalloc(len);
+	if (*url == NULL)
+		return -1;
+
+	return snprintf(*url, len, "http://%s:%d%s", host, port, path);
+}
+#endif /* TRANSPARENT_PROXY */
+
+#ifdef UPSTREAM_SUPPORT
+/*
+ * Add an entry to the upstream list
+ */
+void
+upstream_add(const char *host, int port, const char *domain)
+{
+	char *ptr;
+	struct upstream *up = safemalloc(sizeof (struct upstream));
+
+	if (!up) {
+		log_message(LOG_ERR, "Unable to allocate memory in upstream_add()");
+		return;
+	}
+
+	up->host = up->domain = NULL;
+	up->ip = up->mask = 0;
+
+	if (domain == NULL) {
+		if (!host || host[0] == '\0' || port < 1) {
+			log_message(LOG_WARNING, "Nonsence upstream rule: invalid host or port");
+			goto upstream_cleanup;
+		}
+
+		up->host = safestrdup(host);
+		up->port = port;
+
+		log_message(LOG_INFO, "Added upstream %s:%d for [default]", host, port);
+	} else if (host == NULL) {
+		if (!domain || domain[0] == '\0') {
+			log_message(LOG_WARNING, "Nonsense no-upstream rule: empty domain");
+			goto upstream_cleanup;
+		}
+
+		ptr = strchr(domain, '/');
+		if (ptr) {
+			struct in_addr addrstruct;
+
+			*ptr = '\0';
+			if (inet_aton(domain, &addrstruct) != 0) {
+				up->ip = ntohl(addrstruct.s_addr);
+				*ptr++ = '/';
+
+				if (strchr(ptr, '.')) {
+					if (inet_aton(ptr, &addrstruct) != 0)
+						up->mask = ntohl(addrstruct.s_addr);
+				} else {
+					up->mask = ~((1 << (32 - atoi(ptr))) - 1);
+				}
+			}
+		} else {
+			up->domain = safestrdup(domain);
+		}
+
+		log_message(LOG_INFO, "Added no-upstream for %s", domain);
+	} else {
+		if (!host || host[0] == '\0' || port < 1 || !domain || domain == '\0') {
+			log_message(LOG_WARNING, "Nonsense upstream rule: invalid parameters");
+			goto upstream_cleanup;
+		}
+
+		up->host = safestrdup(host);
+		up->port = port;
+		up->domain = safestrdup(domain);
+
+		log_message(LOG_INFO, "Added upstream %s:%d for %s",
+			    host, port, domain);
+	}
+
+	if (!up->domain && !up->ip) { /* always add default to end */
+		struct upstream *tmp = config.upstream_list;
+
+		while (tmp) {
+			if (!tmp->domain && !tmp->ip) {
+				log_message(LOG_WARNING,
+					    "Duplicate default upstream");
+				goto upstream_cleanup;
+			}
+
+			if (!tmp->next) {
+				up->next = NULL;
+				tmp->next = up;
+				return;
+			}
+
+			tmp = tmp->next;
+		}
+	}
+
+	up->next = config.upstream_list;
+	config.upstream_list = up;
+
+	return;
+
+      upstream_cleanup:
+	safefree(up->host);
+	safefree(up->domain);
+	safefree(up);
+
+	return;
+}
+
+/*
+ * Check if a host is in the upstream list
+ */
+static struct upstream *
+upstream_get(char *host)
+{
+	struct upstream *up = config.upstream_list;
+
+	in_addr_t my_ip = INADDR_NONE;
+
+	while (up) {
+		if (up->domain) {
+			if (strcasecmp(host, up->domain) == 0)
+				break; /* exact match */
+
+			if (up->domain[0] == '.') {
+				char *dot = strchr(host, '.');
+
+				if (!dot && !up->domain[1])
+					break; /* local host matches "." */
+
+				while (dot && strcasecmp(dot, up->domain))
+					dot = strchr(dot+1, '.');
+
+				if (dot)
+					break; /* subdomain match */
+			}
+		} else if (up->ip) {
+			if (my_ip == INADDR_NONE)
+				my_ip = ntohl(inet_addr(host));
+
+			if ((my_ip & up->mask) == up->ip)
+				break;
+		} else {
+			break; /* No domain or IP, default upstream */
+		}
+
+		up = up->next;
+	}
+
+	if (up && (!up->host || !up->port))
+		up = NULL;
+
+	if (up)
+		log_message(LOG_INFO, "Found proxy %s:%d for %s",
+				up->host, up->port, host);
+	else
+		log_message(LOG_INFO, "No proxy for %s", host);
+
+	return up;
+}
+#endif
+
 /*
  * Create a connection for HTTP connections.
  */
 static int
 establish_http_connection(struct conn_s *connptr, struct request_s *request)
 {
+	char portbuff[7];
+
+	/* Build a port string if it's not a standard port */
+	if (request->port != HTTP_PORT && request->port != HTTP_PORT_SSL)
+		snprintf(portbuff, 7, ":%u", request->port);
+	else
+		portbuff[0] = '\0';
+
 	return write_message(connptr->server_fd,
 			     "%s %s HTTP/1.0\r\n" \
-			     "Host: %s\r\n" \
+			     "Host: %s%s\r\n" \
 			     "Connection: close\r\n",
-			     request->method, request->path, request->host);
+			     request->method, request->path,
+			     request->host, portbuff);
 }
 
 /*
@@ -335,13 +546,17 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
 		log_message(LOG_ERR,
 			    "process_request: Bad Request on file descriptor %d",
 			    connptr->client_fd);
-		indicate_http_error(connptr, 400, "Bad Request. No request found.");
+		indicate_http_error(connptr, 400, "Bad Request",
+				    "detail", "Request has an invalid format",
+				    "url", url,
+				    NULL);
 
 		safefree(url);
 		free_request_struct(request);
 
 		return NULL;
 	}
+
 	/* 
 	 * FIXME: We need to add code for the simple HTTP/0.9 style GET
 	 * request.
@@ -351,7 +566,10 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
 		log_message(LOG_ERR,
 			    "process_request: Null URL on file descriptor %d",
 			    connptr->client_fd);
-		indicate_http_error(connptr, 400, "Bad Request. Null URL.");
+		indicate_http_error(connptr, 400, "Bad Request",
+				    "detail", "Request has an empty URL",
+				    "url", url,
+				    NULL);
 
 		safefree(url);
 		free_request_struct(request);
@@ -359,13 +577,15 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
 		return NULL;
 	}
 
-	if (strncasecmp(url, "http://", 7) == 0) {
-		/* Make sure the first four characters are lowercase */
-		memcpy(url, "http", 4);
+	if (strncasecmp(url, "http://", 7) == 0
+	    || (UPSTREAM_CONFIGURED() && strncasecmp(url, "ftp://", 6) == 0)) {
+		char *skipped_type = strstr(url, "//") + 2;
 
-		if (extract_http_url(url, request) < 0) {
-			indicate_http_error(connptr, 400,
-				"Bad Request. Could not parse URL.");
+		if (extract_http_url(skipped_type, request) < 0) {
+			indicate_http_error(connptr, 400, "Bad Request",
+					    "detail", "Could not parse URL",
+					    "url", url,
+					    NULL);
 
 			safefree(url);
 			free_request_struct(request);
@@ -374,8 +594,10 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
 		}
 	} else if (strcmp(request->method, "CONNECT") == 0) {
 		if (extract_ssl_url(url, request) < 0) {
-			indicate_http_error(connptr, 400,
-				"Bad Request. Could not parse URL.");
+			indicate_http_error(connptr, 400, "Bad Request",
+					    "detail", "Could not parse URL",
+					    "url", url,
+					    NULL);
 
 			safefree(url);
 			free_request_struct(request);
@@ -385,8 +607,11 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
 
 		/* Verify that the port in the CONNECT method is allowed */
 		if (check_allowed_connect_ports(request->port) <= 0) {
-			indicate_http_error(connptr, 403,
-				"CONNECT method not allowed with selected port.");
+			indicate_http_error(connptr, 403, "Access violation",
+					    "detail", "The CONNECT method not allowed " \
+					              "with the port you tried to use.",
+					    "url", url,
+					    NULL);
 			log_message(LOG_INFO, "Refused CONNECT method on port %d",
 				    request->port);
 
@@ -417,7 +642,10 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
 				log_message(LOG_ERR,
 					    "process_request: cannot get destination IP for %d",
 					    connptr->client_fd);
-				indicate_http_error(connptr, 400, "Bad Request. Unknown destination.");
+				indicate_http_error(connptr, 400, "Bad Request",
+						    "detail", "Unknown destination",
+						    "url", url,
+						    NULL);
 				safefree(url);
 				free_request_struct(request);
 				return NULL;
@@ -427,27 +655,34 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
 			request->port = ntohs(dest_addr.sin_port);
 			request->path = safemalloc(strlen(url) + 1);
 			strcpy(request->path, url);
+			safefree(url);
+			build_url(&url, request->host, request->port, request->path);
 			log_message(LOG_INFO,
-				    "process_request: trans IP %s http://%s:%d%s for %d",
-				    request->method, request->host, request->port, request->path, connptr->client_fd);
+				    "process_request: trans IP %s %s for %d",
+				    request->method, url, connptr->client_fd);
 		} else {
 			request->host = safemalloc(length+1);
 			if (sscanf(data, "%[^:]:%hu", request->host, &request->port) != 2) {
 				strcpy(request->host, data);
-				request->port = 80;
+				request->port = HTTP_PORT;
 			}
 			request->path = safemalloc(strlen(url) + 1);
 			strcpy(request->path, url);
+			safefree(url);
+			build_url(&url, request->host, request->port, request->path);
 			log_message(LOG_INFO,
-				    "process_request: trans Host %s http://%s:%d%s for %d",
-				    request->method, request->host, request->port, request->path, connptr->client_fd);
+				    "process_request: trans Host %s %s for %d",
+				    request->method, url, connptr->client_fd);
 		}
 		if (config.ipAddr &&
 		    strcmp(request->host, config.ipAddr) == 0) {
 			log_message(LOG_ERR,
 				    "process_request: destination IP is localhost %d",
 				    connptr->client_fd);
-			indicate_http_error(connptr, 400, "Bad Request. Bad destination.");
+			indicate_http_error(connptr, 400, "Bad Request",
+					    "detail", "You tried to connect to the machine the proxy is running on",
+					    "url", url,
+					    NULL);
 			safefree(url);
 			free_request_struct(request);
 			return NULL;
@@ -456,7 +691,10 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
 		log_message(LOG_ERR,
 			    "process_request: Unknown URL type on file descriptor %d",
 			    connptr->client_fd);
-		indicate_http_error(connptr, 400, "Bad Request. Unknown URL type.");
+		indicate_http_error(connptr, 400, "Bad Request",
+				    "detail", "Unknown URL type",
+				    "url", url,
+				    NULL);
 
 		safefree(url);
 		free_request_struct(request);
@@ -487,7 +725,10 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
 					    "Proxying refused on filtered domain \"%s\"",
 					    request->host);
 
-			indicate_http_error(connptr, 403, "Filtered.");
+			indicate_http_error(connptr, 403, "Filtered",
+					    "detail", "The request you made has been filted",
+					    "url", url,
+					    NULL);
 
 			safefree(url);
 			free_request_struct(request);
@@ -529,7 +770,7 @@ process_request(struct conn_s *connptr, hashmap_t hashofheaders)
  *	- rjkaes
  */
 static int
-pull_client_data(struct conn_s *connptr, unsigned long int length)
+pull_client_data(struct conn_s *connptr, long int length)
 {
 	char *buffer;
 	ssize_t len;
@@ -544,7 +785,7 @@ pull_client_data(struct conn_s *connptr, unsigned long int length)
 		if (len <= 0)
 			goto ERROR_EXIT;
 
-		if (!connptr->error_string) {
+		if (!connptr->error_variables) {
 			if (safe_write(connptr->server_fd, buffer, len) < 0)
 				goto ERROR_EXIT;
 		}
@@ -552,6 +793,21 @@ pull_client_data(struct conn_s *connptr, unsigned long int length)
 		length -= len;
 	} while (length > 0);
 
+	/*
+	 * BUG FIX: Internet Explorer will leave two bytes (carriage
+	 * return and line feed) at the end of a POST message.  These
+	 * need to be eaten for tinyproxy to work correctly.
+	 */
+	socket_nonblocking(connptr->client_fd);
+	len = recv(connptr->client_fd, buffer, 2, MSG_PEEK);
+	socket_blocking(connptr->client_fd);
+
+	if (len < 0 && errno != EAGAIN)
+		goto ERROR_EXIT;
+
+	if (len == 2 && CHECK_CRLF(buffer, len))
+		read(connptr->client_fd, buffer, 2);
+
 	safefree(buffer);
 	return 0;
 
@@ -616,7 +872,7 @@ get_all_headers(int fd, hashmap_t hashofheaders)
 {
 	char *header;
 	ssize_t len;
-	bool_t double_cgi = FALSE;
+	unsigned int double_cgi = FALSE; /* boolean */
 
 	assert(fd >= 0);
 	assert(hashofheaders != NULL);
@@ -681,7 +937,7 @@ remove_connection_headers(hashmap_t hashofheaders)
 	ssize_t len;
 	int i;
 
-	for (i = 0; i < (sizeof(headers) / sizeof(char *)); ++i) {
+	for (i = 0; i != (sizeof(headers) / sizeof(char *)); ++i) {
 		/* Look for the connection header.  If it's not found, return. */
 		len = hashmap_entry_by_key(hashofheaders, headers[i], (void **)&data);
 		if (len <= 0)
@@ -735,8 +991,8 @@ get_content_length(hashmap_t hashofheaders)
 }
 
 /*
- * Search for Via head in a hash of headers and either write a new Via header,
- * or append our information to the end of an existing Via header.
+ * Search for Via header in a hash of headers and either write a new Via
+ * header, or append our information to the end of an existing Via header.
  *
  * FIXME: Need to add code to "hide" our internal information for security
  * purposes.
@@ -746,11 +1002,15 @@ write_via_header(int fd, hashmap_t hashofheaders,
 		 unsigned int major, unsigned int minor)
 {
 	ssize_t len;
-	char hostname[128];
+	char hostname[512];
 	char *data;
 	int ret;
 
-	gethostname(hostname, sizeof(hostname));
+	if (config.via_proxy_name) {
+		strlcpy(hostname, config.via_proxy_name, sizeof(hostname));
+	} else if (gethostname(hostname, sizeof(hostname)) < 0) {
+		strcpy(hostname, "unknown");
+	}
 
 	/*
 	 * See if there is a "Via" header.  If so, again we need to do a bit
@@ -802,8 +1062,7 @@ process_client_headers(struct conn_s *connptr, hashmap_t hashofheaders)
 	};
 	int i;
 	hashmap_iter iter;
-	long content_length = -1;
-	int ret;
+	int ret = 0;
 
 	char *data, *header;
 
@@ -822,7 +1081,7 @@ process_client_headers(struct conn_s *connptr, hashmap_t hashofheaders)
 	 * See if there is a "Content-Length" header.  If so, again we need
 	 * to do a bit of processing.
 	 */
-	content_length = get_content_length(hashofheaders);
+	connptr->content_length.client = get_content_length(hashofheaders);
 
 	/*
 	 * See if there is a "Connection" header.  If so, we need to do a bit
@@ -833,7 +1092,7 @@ process_client_headers(struct conn_s *connptr, hashmap_t hashofheaders)
 	/*
 	 * Delete the headers listed in the skipheaders list
 	 */
-	for (i = 0; i < (sizeof(skipheaders) / sizeof(char *)); i++) {
+	for (i = 0; i != (sizeof(skipheaders) / sizeof(char *)); i++) {
 		hashmap_remove(hashofheaders, skipheaders[i]);
 	}
 
@@ -843,7 +1102,9 @@ process_client_headers(struct conn_s *connptr, hashmap_t hashofheaders)
 			       connptr->protocol.minor);
 	if (ret < 0) {
 		indicate_http_error(connptr, 503,
-				    "Could not send data to remote server.");
+				    "Could not send data to remote server",
+				    "detail", "A network error occurred while trying to write data to the remote web server.",
+				    NULL);
 		goto PULL_CLIENT_DATA;
 	}
 
@@ -858,14 +1119,15 @@ process_client_headers(struct conn_s *connptr, hashmap_t hashofheaders)
 					     &data,
 					     (void**)&header);
 
-			if (!is_anonymous_enabled() || anonymous_search(data) <= 0) {
+			if (!is_anonymous_enabled() || anonymous_search(data) > 0) {
 				ret = write_message(connptr->server_fd,
 						    "%s: %s\r\n",
 						    data, header);
 				if (ret < 0) {
-					indicate_http_error(connptr,
-							    503,
-							    "Could not send data to remove server.");
+					indicate_http_error(connptr, 503,
+							    "Could not send data to remote server",
+							    "detail", "A network error occurred while trying to write data to the remote web server.",
+							    NULL);
 					goto PULL_CLIENT_DATA;
 				}
 			}
@@ -885,9 +1147,9 @@ process_client_headers(struct conn_s *connptr, hashmap_t hashofheaders)
 	 * Spin here pulling the data from the client.
 	 */
   PULL_CLIENT_DATA:
-	if (content_length > 0)
+	if (connptr->content_length.client > 0)
 		return pull_client_data(connptr,
-					(unsigned long int) content_length);
+					connptr->content_length.client);
 	else
 		return ret;
 }
@@ -951,7 +1213,9 @@ process_server_headers(struct conn_s *connptr)
 		hashmap_delete(hashofheaders);
 		safefree(response_line);
 
-		indicate_http_error(connptr, 503, "Could not retrieve all the headers from the remote server.");
+		indicate_http_error(connptr, 503, "Could not retrieve all the headers",
+				    "detail", PACKAGE " was unable to retrieve and process headers from the remote web server.",
+				    NULL);
 		return -1;
 	}
 
@@ -965,7 +1229,7 @@ process_server_headers(struct conn_s *connptr)
 	 * If there is a "Content-Length" header, retrieve the information
 	 * from it for later use.
 	 */
-	connptr->remote_content_length = get_content_length(hashofheaders);
+	connptr->content_length.server = get_content_length(hashofheaders);
 
 	/*
 	 * See if there is a connection header.  If so, we need to to a bit of
@@ -976,13 +1240,14 @@ process_server_headers(struct conn_s *connptr)
 	/*
 	 * Delete the headers listed in the skipheaders list
 	 */
-	for (i = 0; i < (sizeof(skipheaders) / sizeof(char *)); i++) {
+	for (i = 0; i != (sizeof(skipheaders) / sizeof(char *)); i++) {
 		hashmap_remove(hashofheaders, skipheaders[i]);
 	}
 
 	/* Send, or add the Via header */
-        ret = write_via_header(connptr->client_fd, hashofheaders,
-			       connptr->protocol.major, connptr->protocol.minor);
+	ret = write_via_header(connptr->client_fd, hashofheaders,
+			       connptr->protocol.major,
+			       connptr->protocol.minor);
 	if (ret < 0)
 		goto ERROR_EXIT;
 
@@ -1088,8 +1353,8 @@ relay_connection(struct conn_s *connptr)
 			if (bytes_received < 0)
 				break;
 
-			connptr->remote_content_length -= bytes_received;
-			if (connptr->remote_content_length == 0)
+			connptr->content_length.server -= bytes_received;
+			if (connptr->content_length.server == 0)
 				break;
 		}
 		if (FD_ISSET(connptr->client_fd, &rset)
@@ -1115,42 +1380,60 @@ relay_connection(struct conn_s *connptr)
 		if (write_buffer(connptr->client_fd, connptr->sbuffer) < 0)
 			break;
 	}
+	shutdown(connptr->client_fd, SHUT_WR);
 
 	/*
 	 * Try to send any remaining data to the server if we can.
 	 */
 	socket_blocking(connptr->server_fd);
 	while (buffer_size(connptr->cbuffer) > 0) {
-		if (write_buffer(connptr->client_fd, connptr->cbuffer) < 0)
+		if (write_buffer(connptr->server_fd, connptr->cbuffer) < 0)
 			break;
 	}
 
 	return;
 }
 
-#ifdef UPSTREAM_SUPPORT
 /*
  * Establish a connection to the upstream proxy server.
  */
 static int
 connect_to_upstream(struct conn_s *connptr, struct request_s *request)
 {
+#ifndef UPSTREAM_SUPPORT
+	/*
+	 * This function does nothing if upstream support was not compiled
+	 * into tinyproxy.
+	 */
+	return -1;
+#else
 	char *combined_string;
 	int len;
 
+	struct upstream *cur_upstream = upstream_get(request->host);
+	if(!cur_upstream) {
+		log_message(LOG_WARNING,
+			    "No upstream proxy defined for %s.",
+			    request->host);
+		indicate_http_error(connptr, 404, "Unable to connect to upstream proxy.");
+		return -1;
+	}
+
 	connptr->server_fd =
-	    opensock(config.upstream_name, config.upstream_port);
+	    opensock(cur_upstream->host, cur_upstream->port);
 
 	if (connptr->server_fd < 0) {
 		log_message(LOG_WARNING,
 			    "Could not connect to upstream proxy.");
-		indicate_http_error(connptr, 404, "Unable to connect to upstream proxy.");
+		indicate_http_error(connptr, 404, "Unable to connect to upstream proxy",
+				    "detail", "A network error occurred while trying to connect to the upstream web proxy.",
+				    NULL);
 		return -1;
 	}
 
 	log_message(LOG_CONN,
 		    "Established connection to upstream proxy \"%s\" using file descriptor %d.",
-		    config.upstream_name, connptr->server_fd);
+		    cur_upstream->host, connptr->server_fd);
 
 	/*
 	 * We need to re-write the "path" part of the request so that we
@@ -1158,9 +1441,9 @@ connect_to_upstream(struct conn_s *connptr, struct request_s *request)
 	 * method and path.
 	 */
 	if (connptr->connect_method) {
-		len = strlen(request->host) + 6;
+		len = strlen(request->host) + 7;
 
-		combined_string = safemalloc(len + 1);
+		combined_string = safemalloc(len);
 		if (!combined_string) {
 			return -1;
 		}
@@ -1169,7 +1452,7 @@ connect_to_upstream(struct conn_s *connptr, struct request_s *request)
 			 request->port);
 	} else {
 		len = strlen(request->host) + strlen(request->path) + 14;
-		combined_string = safemalloc(len + 1);
+		combined_string = safemalloc(len);
 		if (!combined_string) {
 			return -1;
 		}
@@ -1178,71 +1461,13 @@ connect_to_upstream(struct conn_s *connptr, struct request_s *request)
 			 request->port, request->path);
 	}
 
-	safefree(request->path);
+	if (request->path)
+		safefree(request->path);
 	request->path = combined_string;
 
 	return establish_http_connection(connptr, request);
+#endif
 }
-#endif
-
-#ifdef TUNNEL_SUPPORT
-/*
- * If tunnel has been configured then redirect any connections to it.
- */
-static int
-connect_to_tunnel(struct conn_s *connptr)
-{
-
-#if 0
-	/*
-	 * NOTE: This must be fixed
-	 *
-	 * Needed to remove this for right now since it breaks the semantics
-	 * of the "tunnel" concept since the information from the remote host
-	 * wasn't being sent until _after_ data was sent by the client.  This
-	 * is not correct since we should be sending the data regardless of
-	 * who sent it first.
-	 *
-	 * I'll have to look into this for the next release.
-	 */
-	char *request_buf;
-	ssize_t len;
-	int pos;
-
-	request_buf = safemalloc(HTTP_LINE_LENGTH);
-	if (request_buf) {
-		len = recv(connptr->client_fd, request_buf, HTTP_LINE_LENGTH - 1, MSG_PEEK);
-		for (pos = 0; pos < len && request_buf[pos] != '\n'; pos++)
-			;
-		request_buf[pos] = '\0';
-	     
-		log_message(LOG_CONN, "Request: %s", request_buf);
-
-		safefree(request_buf);
-	}
-#endif
-
-	log_message(LOG_INFO, "Redirecting to %s:%d",
-		    config.tunnel_name, config.tunnel_port);
-
-	connptr->server_fd =
-		opensock(config.tunnel_name, config.tunnel_port);
-
-	if (connptr->server_fd < 0) {
-		log_message(LOG_WARNING,
-			    "Could not connect to tunnel.");
-		indicate_http_error(connptr, 404, "Unable to connect to tunnel.");
-
-		return -1;
-	}
-
-	log_message(LOG_INFO,
-		    "Established a connection to the tunnel \"%s\" using file descriptor %d.",
-		    config.tunnel_name, connptr->server_fd);
-
-	return 0;
-}
-#endif
 
 /*
  * This is the main drive for each connection. As you can tell, for the
@@ -1276,25 +1501,19 @@ handle_connection(int fd)
 
 	if (check_acl(fd, peer_ipaddr, peer_string) <= 0) {
 		update_stats(STAT_DENIED);
-		indicate_http_error(connptr, 403,
-			"You do not have authorization for using this service.");
+		indicate_http_error(connptr, 403, "Access denied",
+				    "detail", "The administrator of this proxy has not configured it to service requests from your host.",
+				    NULL);
 		send_http_error_message(connptr);
 		destroy_conn(connptr);
 		return;
 	}
 
-	if (TUNNEL_CONFIGURED()) {
-		if (connect_to_tunnel(connptr) < 0)
-			goto internal_proxy;
-		else
-			goto relay_proxy;
-	}
-
-      internal_proxy:
 	if (read_request_line(connptr) < 0) {
 		update_stats(STAT_BADCONN);
-		indicate_http_error(connptr, 408,
-			"Server timeout waiting for the HTTP request from the client.");
+		indicate_http_error(connptr, 408, "Timeout",
+				    "detail", "Server timeout waiting for the HTTP request from the client.",
+				    NULL);
 		send_http_error_message(connptr);
 		destroy_conn(connptr);
 		return;
@@ -1305,7 +1524,9 @@ handle_connection(int fd)
 	 */
 	if (!(hashofheaders = hashmap_create(HEADER_BUCKETS))) {
 		update_stats(STAT_BADCONN);
-		indicate_http_error(connptr, 503, HTTP503ERROR);
+		indicate_http_error(connptr, 503, "Internal error",
+				    "detail", "An internal server error occurred while processing your request.  Please contact the administrator.",
+				    NULL);
 		send_http_error_message(connptr);
 		destroy_conn(connptr);
 		return;
@@ -1324,7 +1545,7 @@ handle_connection(int fd)
 
 	request = process_request(connptr, hashofheaders);
 	if (!request) {
-		if (!connptr->error_string && !connptr->show_stats) {
+		if (!connptr->error_variables && !connptr->show_stats) {
 			update_stats(STAT_BADCONN);
 			destroy_conn(connptr);
 			hashmap_delete(hashofheaders);
@@ -1333,14 +1554,17 @@ handle_connection(int fd)
 		goto send_error;
 	}
 
-	if (UPSTREAM_CONFIGURED()) {
+	if (UPSTREAM_CONFIGURED() && (UPSTREAM_HOST(request->host) != NULL)) {
 		if (connect_to_upstream(connptr, request) < 0) {
 			goto send_error;
 		}
 	} else {
 		connptr->server_fd = opensock(request->host, request->port);
 		if (connptr->server_fd < 0) {
-			indicate_http_error(connptr, 500, HTTP500ERROR);
+			indicate_http_error(connptr, 500, "Unable to connect",
+					    "detail", PACKAGE " was unable to connect to the remote web server.",
+					    "error", strerror(errno),
+					    NULL);
 			goto send_error;
 		}
 
@@ -1357,7 +1581,7 @@ handle_connection(int fd)
 
 	if (process_client_headers(connptr, hashofheaders) < 0) {
 		update_stats(STAT_BADCONN);
-		if (!connptr->error_string) {
+		if (!connptr->error_variables) {
 			hashmap_delete(hashofheaders);
 			destroy_conn(connptr);
 			return;
@@ -1365,7 +1589,7 @@ handle_connection(int fd)
 	}
 	hashmap_delete(hashofheaders);
 
-	if (connptr->error_string) {
+	if (connptr->error_variables) {
 		send_http_error_message(connptr);
 		destroy_conn(connptr);
 		return;
@@ -1377,7 +1601,7 @@ handle_connection(int fd)
 
 	if (!connptr->connect_method || UPSTREAM_CONFIGURED()) {
 		if (process_server_headers(connptr) < 0) {
-			if (connptr->error_string)
+			if (connptr->error_variables)
 				send_http_error_message(connptr);
 
 			update_stats(STAT_BADCONN);
@@ -1394,7 +1618,6 @@ handle_connection(int fd)
 		}
 	}
 
-      relay_proxy:
 	relay_connection(connptr);
 
 	log_message(LOG_INFO, "Closed connection between local client (fd:%d) and remote client (fd:%d)",
